package net.knarcraft.blacksmith.config.blacksmith;

import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.config.Settings;
import net.knarcraft.blacksmith.config.StargateYamlConfiguration;
import net.knarcraft.blacksmith.util.ConfigHelper;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import net.knarcraft.blacksmith.util.ItemHelper;
import org.bukkit.Material;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.enchantments.Enchantment;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

/**
 * A class which keeps track of all default blacksmith NPC settings and all global blacksmith settings
 */
public class GlobalBlacksmithSettings implements Settings<BlacksmithSetting> {

    private final Map<Material, Double> materialBasePrices = new HashMap<>();
    private final Map<Material, Double> materialPricePerDurabilityPoints = new HashMap<>();
    private final Map<Enchantment, Double> enchantmentCosts = new HashMap<>();
    private final Map<BlacksmithSetting, Object> settings = new HashMap<>();
    private final List<Material> defaultReforgeAbleMaterials = new ArrayList<>();
    private final List<Enchantment> defaultEnchantmentBlockList = new ArrayList<>();
    private final BlacksmithPlugin instance;

    /**
     * Instantiates a new "Global Blacksmith Settings"
     *
     * @param instance <p>The blacksmith plugin to load and save configurations for</p>
     */
    public GlobalBlacksmithSettings(BlacksmithPlugin instance) {
        this.instance = instance;
    }

    /**
     * Loads all configuration values from the config file
     */
    public void load() {
        // Just in case, clear existing values
        this.settings.clear();
        this.materialBasePrices.clear();
        this.materialPricePerDurabilityPoints.clear();
        this.enchantmentCosts.clear();

        // Load/Save settings
        loadSettings();

        // Save any modified values to disk
        instance.saveConfig();
    }

    /**
     * Changes the value of the given setting
     *
     * @param blacksmithSetting <p>The default NPC setting to change</p>
     * @param newValue          <p>The new value for the setting</p>
     */
    public void changeValue(@NotNull BlacksmithSetting blacksmithSetting, @Nullable Object newValue) {
        if (blacksmithSetting.getValueType() == SettingValueType.ENCHANTMENT_LIST ||
                blacksmithSetting.getValueType() == SettingValueType.REFORGE_ABLE_ITEMS) {
            //Workaround to make sure it's treated as the correct type
            this.settings.put(blacksmithSetting, newValue == null ? null : ConfigHelper.asStringList(newValue));
        } else {
            this.settings.put(blacksmithSetting, newValue);
        }
        save();
        if (blacksmithSetting == BlacksmithSetting.REFORGE_ABLE_ITEMS) {
            loadReforgeAbleItems();
        } else if (blacksmithSetting == BlacksmithSetting.ENCHANTMENT_BLOCK_LIST) {
            loadEnchantmentBlockList();
        }
    }

    /**
     * Gets the current raw value of the given default NPC setting
     *
     * @param blacksmithSetting <p>The setting to get</p>
     * @return <p>The current raw setting value</p>
     */
    public @NotNull Object getRawValue(@NotNull BlacksmithSetting blacksmithSetting) {
        return this.settings.get(blacksmithSetting);
    }

    /**
     * Sets the enchantment cost for the given enchantment
     *
     * @param enchantment        <p>The enchantment to set the enchantment cost for</p>
     * @param newEnchantmentCost <p>The new enchantment cost</p>
     */
    public void setEnchantmentCost(Enchantment enchantment, double newEnchantmentCost) {
        if (enchantment == null) {
            if (newEnchantmentCost < 0) {
                throw new IllegalArgumentException("Enchantment cost cannot be negative!");
            }
            this.settings.put(BlacksmithSetting.ENCHANTMENT_COST, newEnchantmentCost);
        } else {
            if (newEnchantmentCost < 0) {
                this.enchantmentCosts.put(enchantment, null);
            } else {
                this.enchantmentCosts.put(enchantment, newEnchantmentCost);
            }
        }
        save();
    }

    /**
     * Sets the price per durability point for the given material
     *
     * @param material <p>The material to set the price per durability point price for</p>
     * @param newPrice <p>The new price per durability point price</p>
     */
    public void setPricePerDurabilityPoint(Material material, double newPrice) {
        if (material == null) {
            if (newPrice < 0) {
                throw new IllegalArgumentException("Price per durability point cannot be negative!");
            }
            this.settings.put(BlacksmithSetting.PRICE_PER_DURABILITY_POINT, newPrice);
        } else {
            //Use a negative price to unset the per-item value
            if (newPrice < 0) {
                this.materialPricePerDurabilityPoints.put(material, null);
            } else {
                this.materialPricePerDurabilityPoints.put(material, newPrice);
            }
        }
        save();
    }

    /**
     * Sets the base price for the given material
     *
     * @param material     <p>The material to set the base price for</p>
     * @param newBasePrice <p>The new base price</p>
     */
    public void setBasePrice(Material material, double newBasePrice) {
        if (material == null) {
            if (newBasePrice < 0) {
                throw new IllegalArgumentException("Base price cannot be negative!");
            }
            settings.put(BlacksmithSetting.BASE_PRICE, newBasePrice);
        } else {
            //Use a negative price to unset the per-item value
            if (newBasePrice < 0) {
                this.materialBasePrices.put(material, null);
            } else {
                this.materialBasePrices.put(material, newBasePrice);
            }
        }
        save();
    }

    /**
     * Gets the current value of the default NPC settings
     *
     * @return <p>The current value of the default NPC settings</p>
     */
    public Map<BlacksmithSetting, Object> getDefaultNPCSettings() {
        return new HashMap<>(this.settings);
    }

    /**
     * Gets whether to use natural cost for cost calculation
     *
     * <p>Natural cost makes it more costly the more damage is dealt to an item. The alternative is the legacy behavior
     * where the amount of durability points remaining increases the cost.</p>
     *
     * @return <p>Whether to use natural cost</p>
     */
    public boolean getUseNaturalCost() {
        return asBoolean(BlacksmithSetting.NATURAL_COST);
    }

    /**
     * Gets whether to show exact time for reforging wait-time, and for wait-time between sessions
     *
     * @return <p>Whether to show exact time</p>
     */
    public boolean getShowExactTime() {
        return asBoolean(BlacksmithSetting.SHOW_EXACT_TIME);
    }

    /**
     * Gets the base price for the given material
     *
     * @param material <p>The material to get the base price for</p>
     * @return <p>The base price for the material</p>
     */
    public double getBasePrice(Material material) {
        if (this.materialBasePrices.containsKey(material) && this.materialBasePrices.get(material) != null) {
            return this.materialBasePrices.get(material);
        } else {
            return asDouble(BlacksmithSetting.BASE_PRICE);
        }
    }

    /**
     * Gets the price per durability point for the given material
     *
     * @param material <p>The material to get the durability point price for</p>
     * @return <p>The durability point price for the material</p>
     */
    public double getPricePerDurabilityPoint(Material material) {
        if (this.materialPricePerDurabilityPoints.containsKey(material) &&
                this.materialPricePerDurabilityPoints.get(material) != null) {
            return this.materialPricePerDurabilityPoints.get(material);
        } else {
            return asDouble(BlacksmithSetting.PRICE_PER_DURABILITY_POINT);
        }
    }

    /**
     * Gets the cost to be added for each level of the given enchantment
     *
     * @param enchantment <p>The enchantment to get the cost for</p>
     * @return <p>The cost of each enchantment level</p>
     */
    public double getEnchantmentCost(Enchantment enchantment) {
        if (this.enchantmentCosts.containsKey(enchantment) && this.enchantmentCosts.get(enchantment) != null) {
            return this.enchantmentCosts.get(enchantment);
        } else {
            return asDouble(BlacksmithSetting.ENCHANTMENT_COST);
        }
    }

    /**
     * Gets the value of reforgeAbleItems
     *
     * @return <p>The value of reforgeAbleItems</p>
     */
    public @NotNull List<Material> getReforgeAbleItems() {
        return this.defaultReforgeAbleMaterials;
    }

    /**
     * Gets the value of enchantmentBlockList
     *
     * @return <p>The value of enchantmentBlockList</p>
     */
    public @NotNull List<Enchantment> getEnchantmentBlockList() {
        return this.defaultEnchantmentBlockList;
    }

    /**
     * Gets the cost for repairing the given type of anvil
     *
     * @param material <p>The anvil material to repair</p>
     * @return <p>The cost of repairing the anvil</p>
     */
    public double getAnvilCost(Material material) {
        if (material == Material.CHIPPED_ANVIL) {
            return asDouble(BlacksmithSetting.ANVIL_CHIPPED_COST);
        } else if (material == Material.DAMAGED_ANVIL) {
            return asDouble(BlacksmithSetting.ANVIL_DAMAGED_COST);
        } else {
            throw new IllegalArgumentException("An unexpected item was encountered!");
        }
    }

    /**
     * Gets the given value as a boolean
     *
     * <p>This will throw an exception if used for a non-boolean value</p>
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a boolean</p>
     */
    public boolean asBoolean(BlacksmithSetting setting) {
        return ConfigHelper.asBoolean(getValue(setting));
    }

    /**
     * Gets the given value as a double
     *
     * <p>This will throw an exception if used for a non-double setting</p>
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a double</p>
     */
    public double asDouble(BlacksmithSetting setting) {
        return ConfigHelper.asDouble(getValue(setting));
    }

    /**
     * Gets the value of a setting, using the default if not set
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The current value</p>
     */
    private Object getValue(BlacksmithSetting setting) {
        Object value = this.settings.get(setting);
        //If not set in config.yml, use the default value from the enum
        if (value == null) {
            value = setting.getDefaultValue();
        }
        return value;
    }

    /**
     * Loads all global settings
     */
    private void loadSettings() {
        instance.reloadConfig();
        FileConfiguration configuration = instance.getConfiguration();

        for (BlacksmithSetting blacksmithSetting : BlacksmithSetting.values()) {
            if (!configuration.contains(blacksmithSetting.getPath())) {
                //If the setting does not exist in the config file, add it
                configuration.set(blacksmithSetting.getPath(), blacksmithSetting.getDefaultValue());
            } else {
                //Set the setting to the value found in the path
                settings.put(blacksmithSetting, configuration.get(blacksmithSetting.getPath()));
            }
        }
        loadReforgeAbleItems();
        loadEnchantmentBlockList();

        //Load all base prices
        loadBasePrices(configuration);

        //Load all per-durability-point prices
        loadPricesPerDurabilityPoint(configuration);

        //Load all enchantment prices
        loadEnchantmentPrices(configuration);
    }

    /**
     * Loads all prices for enchantments
     *
     * @param fileConfiguration <p>The configuration to read</p>
     */
    private void loadEnchantmentPrices(@NotNull FileConfiguration fileConfiguration) {
        ConfigurationSection enchantmentCostNode = fileConfiguration.getConfigurationSection(
                getBase(BlacksmithSetting.ENCHANTMENT_COST.getPath()));
        if (enchantmentCostNode == null) {
            instance.getLogger().log(Level.WARNING, "Could not load enchantment prices. because the " +
                    "configuration section doesn't exist");
            return;
        }
        Map<String, String> relevantKeys = getRelevantKeys(enchantmentCostNode);
        for (String key : relevantKeys.keySet()) {
            String enchantmentName = relevantKeys.get(key);
            Enchantment enchantment = InputParsingHelper.matchEnchantment(enchantmentName);
            instance.getLogger().log(Level.WARNING, "loadEnchantmentPrices " + enchantmentName);
            setItemPrice(this.enchantmentCosts, enchantmentName, enchantment, enchantmentCostNode.getDouble(key));
        }
    }

    /**
     * Loads all prices per durability point for all materials
     *
     * @param fileConfiguration <p>The configuration to read</p>
     */
    private void loadPricesPerDurabilityPoint(@NotNull FileConfiguration fileConfiguration) {
        ConfigurationSection basePerDurabilityPriceNode = fileConfiguration.getConfigurationSection(
                getBase(BlacksmithSetting.PRICE_PER_DURABILITY_POINT.getPath()));
        if (basePerDurabilityPriceNode == null) {
            instance.getLogger().log(Level.WARNING, "Could not load per durability prices. because the " +
                    "configuration section doesn't exist");
            return;
        }
        Map<String, String> relevantKeys = getRelevantKeys(basePerDurabilityPriceNode);

        for (String key : relevantKeys.keySet()) {
            String materialName = relevantKeys.get(key);
            double price = basePerDurabilityPriceNode.getDouble(key);
            if (materialName.contains("*")) {
                //Treat *CHESTPLATE as a regular expression to match all chest-plates
                setMatchedMaterialPrices(this.materialPricePerDurabilityPoints, materialName, price);
            } else {
                Material material = InputParsingHelper.matchMaterial(materialName);
                setItemPrice(this.materialPricePerDurabilityPoints, materialName, material, price);
            }
        }
    }

    /**
     * Loads base prices for all materials
     *
     * @param fileConfiguration <p>The configuration to read</p>
     */
    private void loadBasePrices(@NotNull FileConfiguration fileConfiguration) {
        ConfigurationSection basePriceNode = fileConfiguration.getConfigurationSection(
                getBase(BlacksmithSetting.BASE_PRICE.getPath()));
        if (basePriceNode == null) {
            instance.getLogger().log(Level.WARNING, "Could not load base prices, because the configuration " +
                    "section doesn't exist");
            return;
        }
        Map<String, String> relevantKeys = getRelevantKeys(basePriceNode);

        for (String key : relevantKeys.keySet()) {
            String materialName = relevantKeys.get(key);
            double price = basePriceNode.getDouble(key);
            if (materialName.contains("*")) {
                //Treat *CHESTPLATE as a regular expression to match all chest-plates
                setMatchedMaterialPrices(this.materialBasePrices, materialName, price);
            } else {
                Material material = InputParsingHelper.matchMaterial(materialName);
                setItemPrice(this.materialBasePrices, materialName, material, price);
            }
        }
    }

    /**
     * Sets the price for any materials matching the given wildcard material name
     *
     * @param prices       <p>The map to store the prices in</p>
     * @param materialName <p>The material name to match</p>
     * @param price        <p>The price to set for the matched materials</p>
     */
    private void setMatchedMaterialPrices(@NotNull Map<Material, Double> prices, @NotNull String materialName, double price) {
        String search = InputParsingHelper.regExIfy(materialName);
        for (Material material : ItemHelper.getAllReforgeAbleMaterials()) {
            if (material.name().matches(search)) {
                setItemPrice(prices, material.name(), material, price);
            }
        }
    }

    /**
     * Sets the price for the given material
     *
     * @param prices   <p>The map to store the price in</p>
     * @param itemName <p>The name of the material to add a price for</p>
     * @param item     <p>The material parsed from the name</p>
     * @param price    <p>The price to set</p>
     */
    private <K> void setItemPrice(@NotNull Map<K, Double> prices, @NotNull String itemName, @Nullable K item,
                                  double price) {
        if (item != null) {
            prices.put(item, price);
        } else {
            instance.getLogger().log(Level.WARNING, "Unable to find a material/enchantment matching " + itemName);
        }
    }

    /**
     * Gets a map between relevant keys and their normalized name
     *
     * @param configurationSection <p>The configuration section to search</p>
     * @return <p>Any sub-keys found that aren't the default</p>
     */
    @NotNull
    private Map<String, String> getRelevantKeys(@NotNull ConfigurationSection configurationSection) {
        Map<String, String> relevant = new HashMap<>();
        for (String dataKey : StargateYamlConfiguration.getKeysWithoutComments(configurationSection, false)) {
            //Skip the default value
            if (dataKey.equals("default")) {
                continue;
            }
            String normalizedName = dataKey.toUpperCase().replace("-", "_");
            relevant.put(dataKey, normalizedName);
        }
        return relevant;
    }

    /**
     * Converts a normalized material name to the format used in the config file
     *
     * @param normalizedName <p>The normalized name to un-normalize</p>
     * @return <p>The un-normalized name</p>
     */
    @NotNull
    private String unNormalizeName(@NotNull String normalizedName) {
        return normalizedName.toLowerCase().replace("_", "-");
    }

    /**
     * Loads reforgeAble items from the current value
     */
    private void loadReforgeAbleItems() {
        this.defaultReforgeAbleMaterials.clear();
        List<String> materialNames = ConfigHelper.asStringList(settings.get(
                BlacksmithSetting.REFORGE_ABLE_ITEMS));
        if (materialNames != null) {
            this.defaultReforgeAbleMaterials.addAll(ItemHelper.getItems(materialNames, true));
        }
    }

    /**
     * Loads the enchantment block list from the current value
     */
    private void loadEnchantmentBlockList() {
        this.defaultEnchantmentBlockList.clear();
        List<String> enchantmentNames = ConfigHelper.asStringList(settings.get(
                BlacksmithSetting.ENCHANTMENT_BLOCK_LIST));
        if (enchantmentNames != null) {
            this.defaultEnchantmentBlockList.addAll(BlacksmithNPCSettings.getEnchantmentBlockList(enchantmentNames));
        }
    }

    /**
     * Saves all current settings to the config file
     */
    private void save() {
        FileConfiguration fileConfiguration = instance.getConfiguration();

        //Save all default settings
        for (BlacksmithSetting setting : BlacksmithSetting.values()) {
            fileConfiguration.set(setting.getPath(), this.settings.get(setting));
        }

        //Save all base prices
        ConfigurationSection basePriceNode = getAndCreateSection(fileConfiguration, getBase(BlacksmithSetting.BASE_PRICE.getPath()));
        for (Material material : this.materialBasePrices.keySet()) {
            basePriceNode.set(unNormalizeName(material.name()), this.materialBasePrices.get(material));
        }

        //Save all per-durability-point prices
        ConfigurationSection basePerDurabilityPriceNode = getAndCreateSection(fileConfiguration, getBase(BlacksmithSetting.PRICE_PER_DURABILITY_POINT.getPath()));
        for (Material material : this.materialPricePerDurabilityPoints.keySet()) {
            basePerDurabilityPriceNode.set(unNormalizeName(material.name()), this.materialPricePerDurabilityPoints.get(material));
        }

        //Load all enchantment prices
        ConfigurationSection enchantmentCostNode = getAndCreateSection(fileConfiguration, getBase(BlacksmithSetting.ENCHANTMENT_COST.getPath()));
        for (Enchantment enchantment : this.enchantmentCosts.keySet()) {
            enchantmentCostNode.set(unNormalizeName(enchantment.getKey().getKey()), this.enchantmentCosts.get(enchantment));
        }

        //Perform the actual save to disk
        instance.saveConfig();
    }

    /**
     * Gets a configuration section, creating it if necessary
     *
     * @param fileConfiguration <p>The file configuration to read</p>
     * @param sectionPath       <p>The path to the configuration section to get</p>
     * @return <p>The configuration section</p>
     */
    @NotNull
    private ConfigurationSection getAndCreateSection(@NotNull FileConfiguration fileConfiguration, @NotNull String sectionPath) {
        ConfigurationSection configurationSection = fileConfiguration.getConfigurationSection(sectionPath);
        if (configurationSection == null) {
            fileConfiguration.createSection(sectionPath);
            configurationSection = fileConfiguration.getConfigurationSection(sectionPath);
        }
        if (configurationSection == null) {
            throw new RuntimeException("Unable to create configuration section!");
        }
        return configurationSection;
    }

    /**
     * Gets the base path of a configuration path pointing to a specific value
     *
     * @param configurationPath <p>The configuration path to get the base of</p>
     * @return <p>The base configuration path</p>
     */
    @NotNull
    private String getBase(@NotNull String configurationPath) {
        String[] parts = configurationPath.split("\\.");
        String[] base = Arrays.copyOfRange(parts, 0, parts.length - 1);
        return String.join(".", base);
    }

}
