package net.knarcraft.blacksmith.trait;

import net.citizensnpcs.api.util.DataKey;
import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.SalvageMethod;
import net.knarcraft.blacksmith.config.SmithPreset;
import net.knarcraft.blacksmith.config.SmithPresetFilter;
import net.knarcraft.blacksmith.config.scrapper.ScrapperNPCSettings;
import net.knarcraft.blacksmith.config.scrapper.ScrapperSetting;
import net.knarcraft.blacksmith.manager.EconomyManager;
import net.knarcraft.blacksmith.util.ItemHelper;
import net.knarcraft.blacksmith.util.SalvageHelper;
import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.formatting.StringReplacer;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ArmorMeta;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static net.knarcraft.blacksmith.formatting.BlacksmithStringFormatter.sendNPCMessage;

/**
 * The class representing a scrapper NPC trait
 */
public class ScrapperTrait extends CustomTrait<ScrapperSetting> {

    private final ScrapperNPCSettings config;

    /**
     * Instantiates a new blacksmith trait
     */
    public ScrapperTrait() {
        super("scrapper");
        //This should crash if the blacksmith plugin hasn't been properly registered
        Bukkit.getServer().getPluginManager().getPlugin("Blacksmith");
        this.config = new ScrapperNPCSettings(BlacksmithPlugin.getInstance().getGlobalScrapperSettings());
        super.setTraitSettings(this.config);
    }

    /**
     * Gets the current settings for this NPC
     *
     * @return <p>The current settings for this NPC</p>
     */
    @NotNull
    public ScrapperNPCSettings getSettings() {
        return config;
    }

    /**
     * Loads all config values stored in citizens' config file for this NPC
     *
     * @param key <p>The data key used for the config root</p>
     */
    @Override
    public void load(@NotNull DataKey key) {
        getSettings().loadVariables(key);
    }

    /**
     * Saves all config values for this NPC
     *
     * @param key <p>The data key used for the config root</p>
     */
    @Override
    public void save(@NotNull DataKey key) {
        getSettings().saveVariables(key);
    }

    /**
     * Starts a new session, and prepares to repair the player's item
     *
     * @param player <p>The player to start the session for</p>
     */
    public void startSession(@NotNull Player player) {
        ItemStack itemInHand = player.getInventory().getItemInMainHand().clone();
        List<Material> salvageAbleItems = getSettings().getSalvageAbleItems();
        boolean extended = getSettings().extendedSalvageEnabled();

        // Check if the item can be salvaged
        if (!canBeSalvaged(itemInHand, salvageAbleItems, extended)) {
            sendNPCMessage(this.npc, player, StringFormatter.replacePlaceholder(getSettings().getInvalidItemMessage(),
                    "{title}", getSettings().getScrapperTitle()));
            return;
        }

        List<ItemStack> salvage = null;

        // Deal with armor trim salvage
        SalvageMethod salvageMethod = SalvageMethod.SALVAGE;
        if (itemInHand.getItemMeta() instanceof ArmorMeta armorMeta && armorMeta.hasTrim()) {
            if (!getSettings().salvageArmorTrims()) {
                sendNPCMessage(this.npc, player, getSettings().getCannotSalvageArmorTrimMessage());
                return;
            }
            salvage = SalvageHelper.getArmorTrimSalvage(itemInHand, armorMeta);
            if (salvage == null) {
                sendNPCMessage(this.npc, player, getSettings().getArmorTrimSalvageNotFoundMessage());
                return;
            }
            salvageMethod = SalvageMethod.ARMOR_TRIM;
        }

        // Remove the netherite ingot from the item
        if (salvage == null && SmithPreset.BLACKSMITH.getFilteredMaterials(SmithPresetFilter.NETHERITE).contains(itemInHand.getType())) {
            if (!getSettings().salvageNetherite()) {
                sendNPCMessage(this.npc, player, getSettings().getCannotSalvageNetheriteMessage());
                return;
            }

            salvage = SalvageHelper.getNetheriteSalvage(itemInHand);
            salvageMethod = SalvageMethod.NETHERITE;
        }

        // As there is no recipe for netherite items, the check needs to be after the netherite salvage check
        if (salvageMethod == SalvageMethod.SALVAGE && !SalvageHelper.isSalvageable(player.getServer(), itemInHand)) {
            sendNPCMessage(this.npc, player, StringFormatter.replacePlaceholder(getSettings().getInvalidItemMessage(),
                    "{title}", getSettings().getScrapperTitle()));
            return;
        }

        // Check if the item is enchanted, and whether this blacksmith can salvage it
        if (!itemInHand.getEnchantments().isEmpty() && !getSettings().salvageEnchanted()) {
            sendNPCMessage(this.npc, player, getSettings().getCannotSalvageEnchantedMessage());
            return;
        }

        // Check if any salvage will be produced
        if (salvage == null) {
            salvage = getSalvage(itemInHand, extended);
        }
        boolean noUsefulSalvage = salvage == null || salvage.isEmpty();
        if (noUsefulSalvage) {
            sendNPCMessage(this.npc, player, getSettings().getTooDamagedMessage());
            return;
        }

        //Start a new scrapper session for the player
        currentSessionStartTime = System.currentTimeMillis();
        int itemsConsumed = SalvageHelper.getRequiredAmountForSalvage(player.getServer(), itemInHand);
        session = new SalvageSession(this, player, npc, getSettings(), salvage, salvageMethod, itemsConsumed);
        // Print the cost to the player
        printCostMessage(player, itemInHand, EconomyManager.formatSalvageCost(salvageMethod), salvageMethod);
    }

    /**
     * Prints a message to the given player, explaining the cost of salvaging the held item
     *
     * @param player        <p>The player that interacted with the scrapper</p>
     * @param itemInHand    <p>The item the player wants to salvage</p>
     * @param cost          <p>The cost of salvaging the item</p>
     * @param salvageMethod <p>The type of salvage performed</p>
     */
    private void printCostMessage(@NotNull Player player, @NotNull ItemStack itemInHand, @NotNull String cost,
                                  @NotNull SalvageMethod salvageMethod) {
        StringReplacer replacer = new StringReplacer();
        replacer.add("{cost}", cost);
        replacer.add("{item}", itemInHand.getType().name().toLowerCase().replace('_', ' '));
        if (salvageMethod == SalvageMethod.ARMOR_TRIM) {
            sendNPCMessage(this.npc, player, replacer.replace(getSettings().getArmorTrimCostMessage()));
        } else if (salvageMethod == SalvageMethod.SALVAGE) {
            String expectedYield;
            if (ItemHelper.getDamage(itemInHand) <= 0) {
                expectedYield = getSettings().getFullSalvageMessage();
            } else {
                expectedYield = getSettings().getPartialSalvageMessage();
            }
            replacer.add("{yield}", expectedYield);
            sendNPCMessage(this.npc, player, replacer.replace(getSettings().getCostMessage()));
        } else if (salvageMethod == SalvageMethod.NETHERITE) {
            sendNPCMessage(this.npc, player, replacer.replace(getSettings().getNetheriteCostMessage()));
        }
    }

    @Override
    protected boolean showExactTime() {
        return BlacksmithPlugin.getInstance().getGlobalScrapperSettings().showExactTime();
    }

    /**
     * Gets whether this scrapper can salvage the given item
     *
     * @param item             <p>The item to check</p>
     * @param salvageAbleItems <p>The items this scrapper can salvage</p>
     * @param extended         <p>Whether extended salvage is enabled</p>
     * @return <p>True if the item can be theoretically salvaged</p>
     */
    private boolean canBeSalvaged(@NotNull ItemStack item, @NotNull List<Material> salvageAbleItems, boolean extended) {
        return (extended || ItemHelper.isRepairable(item)) &&
                (salvageAbleItems.isEmpty() || salvageAbleItems.contains(item.getType()));
    }

    /**
     * Gets salvage for an item, if it's salvage-able
     *
     * @param item     <p>The item to calculate salvage for</p>
     * @param extended <p>Whether extended salvage is enabled</p>
     * @return <p>The possible salvage, or null if not salvage-able</p>
     */
    @Nullable
    private List<ItemStack> getSalvage(@NotNull ItemStack item, boolean extended) {
        // Get the salvage, for the item, but ignore some materials if set, and the item isn't at full durability
        Set<Material> trashSalvage = BlacksmithPlugin.getInstance().getGlobalScrapperSettings().getTrashSalvage(
                item.getType());
        // Don't ignore salvage for fully repaired items
        if (trashSalvage == null || ItemHelper.getDamage(item) == 0) {
            trashSalvage = new HashSet<>();
        }

        return SalvageHelper.getSalvage(BlacksmithPlugin.getInstance().getServer(), item, trashSalvage, extended);
    }

}
