package net.knarcraft.blacksmith.util;

import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.formatting.BlacksmithTranslatableMessage;
import org.bukkit.Material;
import org.bukkit.command.CommandSender;
import org.bukkit.enchantments.Enchantment;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;

/**
 * A helper class for validating a value's type
 */
public final class TypeValidationHelper {

    private TypeValidationHelper() {

    }

    /**
     * Gets whether the given value is valid for the given value type
     *
     * @param valueType <p>The value type required</p>
     * @param value     <p>The value given</p>
     * @param sender    <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is valid</p>
     */
    public static boolean isValid(@NotNull SettingValueType valueType, @Nullable Object value,
                                  @Nullable CommandSender sender) {
        try {
            return switch (valueType) {
                case POSITIVE_DOUBLE -> isPositiveDouble(value, sender);
                case STRING -> isNonEmptyString(value, sender);
                case POSITIVE_INTEGER -> isPositiveInteger(value, sender);
                case PERCENTAGE -> isPercentage(value, sender);
                case BOOLEAN -> true;
                case REFORGE_ABLE_ITEMS, STRING_LIST -> isStringList(value, sender);
                case MATERIAL -> isMaterial(value, sender);
                case ENCHANTMENT -> isEnchantment(value, sender);
                case ENCHANTMENT_LIST -> isEnchantmentList(value, sender);
            };
        } catch (ClassCastException exception) {
            //This error signifies that an object is not a string, and of the wrong class
            return false;
        }
    }

    /**
     * Checks whether the given value is a material
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is a material</p>
     */
    private static boolean isMaterial(@Nullable Object value, @Nullable CommandSender sender) {
        boolean isMaterial = value instanceof Material || (value instanceof String string &&
                InputParsingHelper.matchMaterial(string) != null);
        if (!isMaterial && sender != null) {
            BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                    BlacksmithTranslatableMessage.ITEM_TYPE_MATERIAL);
        }
        return isMaterial;
    }

    /**
     * Checks whether the given value is a list of enchantments
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is an enchantment list</p>
     */
    private static boolean isEnchantmentList(@Nullable Object value, @Nullable CommandSender sender) {
        // Check whether a string list is given
        if (!isStringList(value, sender)) {
            return false;
        }

        // Make sure the input can be converted to a string list
        List<String> strings = ConfigHelper.asStringList(value);
        if (strings == null) {
            return false;
        }

        // Make sure each value is an enchantment
        for (String string : strings) {
            if (!isEnchantment(string, sender)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Checks whether the given value is an enchantment
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is an enchantment</p>
     */
    private static boolean isEnchantment(@Nullable Object value, @Nullable CommandSender sender) {
        boolean isEnchantment = value instanceof Enchantment || (value instanceof String string &&
                InputParsingHelper.matchEnchantment(string) != null);
        if (!isEnchantment && sender != null) {
            BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                    BlacksmithTranslatableMessage.ITEM_TYPE_ENCHANTMENT);
        }
        return isEnchantment;
    }

    /**
     * Checks whether the given value is a string list
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is a string list</p>
     */
    private static boolean isStringList(@Nullable Object value, @Nullable CommandSender sender) {
        boolean isStringList = value instanceof String[] || value instanceof List<?> || value instanceof String;
        if (!isStringList && sender != null) {
            BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                    BlacksmithTranslatableMessage.INPUT_STRING_LIST_REQUIRED);
        }
        return isStringList;
    }

    /**
     * Checks whether the given value is a valid percentage
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is a percentage</p>
     */
    private static boolean isPercentage(@Nullable Object value, @Nullable CommandSender sender) {
        boolean isPercentage = value != null && isPositiveInteger(value, null) &&
                ConfigHelper.asInt(value) >= 0 && ConfigHelper.asInt(value) <= 100;
        if (!isPercentage && sender != null) {
            BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                    BlacksmithTranslatableMessage.INPUT_PERCENTAGE_REQUIRED);
        }
        return isPercentage;
    }

    /**
     * Checks whether the given value is a non-empty string
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is a non-empty string</p>
     */
    private static boolean isNonEmptyString(@Nullable Object value, @Nullable CommandSender sender) {
        boolean isString = value instanceof String string && !string.isBlank();
        if (!isString && sender != null) {
            BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                    BlacksmithTranslatableMessage.INPUT_STRING_REQUIRED);
        }
        return isString;
    }

    /**
     * Checks whether the given value is a positive double
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is a positive double</p>
     */
    private static boolean isPositiveDouble(@Nullable Object value, @Nullable CommandSender sender) {
        try {
            return value != null && ConfigHelper.asDouble(value) >= 0.0;
        } catch (NumberFormatException | NullPointerException exception) {
            if (sender != null) {
                BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                        BlacksmithTranslatableMessage.INPUT_POSITIVE_DOUBLE_REQUIRED);
            }
            return false;
        }
    }

    /**
     * Checks whether the given value is a positive integer
     *
     * @param value  <p>The value to check</p>
     * @param sender <p>The command sender to use for printing error messages</p>
     * @return <p>True if the value is a positive integer</p>
     */
    private static boolean isPositiveInteger(@Nullable Object value, @Nullable CommandSender sender) {
        try {
            return value != null && ConfigHelper.asInt(value) >= 0;
        } catch (NumberFormatException | NullPointerException exception) {
            if (sender != null) {
                BlacksmithPlugin.getStringFormatter().displayErrorMessage(sender,
                        BlacksmithTranslatableMessage.INPUT_POSITIVE_INTEGER_REQUIRED);
            }
            return false;
        }
    }

}
